from typing import List
from pathlib import Path

from spotdl.search import SongObject, song_gatherer
from spotdl.providers import provider_utils, metadata_provider


def parse_query(
    query: List[str], format, use_youtube, generate_m3u, threads
) -> List[SongObject]:
    """
    Parse query and return list containing song object
    """

    songs_list = []

    # Iterate over all search queries and add them to songs_list
    for request in query:
        if request.endswith(".spotdlTrackingFile"):
            continue

        songs_list.extend(
            parse_request(request, format, use_youtube, generate_m3u, threads)
        )

        # linefeed to visually separate output for each query
        print()

    return songs_list


def parse_request(
    request: str,
    output_format: str = None,
    use_youtube: bool = False,
    generate_m3u: bool = False,
    threads: int = 1,
) -> List[SongObject]:
    song_list: List[SongObject] = []
    if (
        "youtube.com/watch?v=" in request
        and "open.spotify.com" in request
        and "track" in request
        and "|" in request
    ):
        urls = request.split("|")

        if len(urls) <= 1 or "youtube" not in urls[0] or "spotify" not in urls[1]:
            print("Incorrect format used, please use YouTubeURL|SpotifyURL")
        else:
            print("Fetching YouTube video with spotify metadata")
            song_list = [
                song
                for song in [get_youtube_meta_track(urls[0], urls[1], output_format)]
                if song is not None
            ]
    elif "open.spotify.com" in request and "track" in request:
        print("Fetching Song...")
        song = song_gatherer.from_spotify_url(request, output_format, use_youtube)
        try:
            song_list = [song] if song.youtube_link is not None else []
        except (OSError, ValueError, LookupError):
            song_list = []
    elif "open.spotify.com" in request and "album" in request:
        print("Fetching Album...")
        song_list = song_gatherer.from_album(
            request, output_format, use_youtube, threads
        )
    elif "open.spotify.com" in request and "playlist" in request:
        print("Fetching Playlist...")
        song_list = song_gatherer.from_playlist(
            request, output_format, use_youtube, generate_m3u, threads
        )
    elif "open.spotify.com" in request and "artist" in request:
        print("Fetching artist...")
        song_list = song_gatherer.from_artist(
            request, output_format, use_youtube, threads
        )
    elif request == "saved":
        print("Fetching Saved Songs...")
        song_list = song_gatherer.from_saved_tracks(output_format, use_youtube, threads)
    else:
        print('Searching Spotify for song named "%s"...' % request)
        try:
            song_list = song_gatherer.from_search_term(
                request, output_format, use_youtube
            )
        except Exception as e:
            print(e)

    # filter out NONE songObj items (already downloaded)
    song_list = [song_object for song_object in song_list if song_object is not None]

    return song_list


def get_youtube_meta_track(
    youtube_url: str, spotify_url: str, output_format: str = None
):
    # check if URL is a playlist, user, artist or album, if yes raise an Exception,
    # else procede

    # Get the Song Metadata
    print(f"Gathering Spotify Metadata for: {spotify_url}")
    raw_track_meta, raw_artist_meta, raw_album_meta = metadata_provider.from_url(
        spotify_url
    )

    song_name = raw_track_meta["name"]
    contributing_artist = []
    for artist in raw_track_meta["artists"]:
        contributing_artist.append(artist["name"])

    converted_file_name = SongObject.create_file_name(
        song_name, [artist["name"] for artist in raw_track_meta["artists"]]
    )

    if output_format is None:
        output_format = "mp3"

    converted_file_path = Path(".", f"{converted_file_name}.{output_format}")

    # if a song is already downloaded skip it
    if converted_file_path.is_file():
        print(f'Skipping "{converted_file_name}" as it\'s already downloaded')
        return None

    # (try to) Get lyrics from Genius
    lyrics = provider_utils._get_song_lyrics(song_name, contributing_artist)

    return SongObject(
        raw_track_meta, raw_album_meta, raw_artist_meta, youtube_url, lyrics
    )



from spotdl.providers.provider_utils import _parse_duration


def test_parse_duration():
    """
    Test the duration parsing
    """

    assert _parse_duration("3:16") == float(196.0)  # 3 min song
    assert _parse_duration("20") == float(20.0)  # 20 second song
    assert _parse_duration("25:59") == float(1559.0)  # 26 min result
    assert _parse_duration("25:59:59") == float(93599.0)  # 26 hour result
    assert _parse_duration("likes") == float(0.0)  # bad values
    assert _parse_duration("views") == float(0.0)
    assert _parse_duration([1, 2, 3]) == float(0.0)  # type: ignore
    assert _parse_duration({"json": "data"}) == float(0.0)  # type: ignore



from typing import Optional

from rich.text import Text
from rich.theme import Theme
from rich.progress import Task
from rich.console import Console
from rich.style import StyleType
from rich.highlighter import Highlighter
from rich.progress import BarColumn, TimeRemainingColumn, Progress, ProgressColumn
from rich.console import (
    JustifyMethod,
    detect_legacy_windows,
    OverflowMethod,
)

from spotdl.search import SongObject

custom_theme = Theme(
    {
        "bar.back": "grey23",
        "bar.complete": "rgb(165,66,129)",
        "bar.finished": "rgb(114,156,31)",
        "bar.pulse": "rgb(165,66,129)",
        "general": "green",
        "nonimportant": "rgb(40,100,40)",
        "progress.data.speed": "red",
        "progress.description": "none",
        "progress.download": "green",
        "progress.filesize": "green",
        "progress.filesize.total": "green",
        "progress.percentage": "green",
        "progress.remaining": "rgb(40,100,40)",
    }
)


class YTDLLogger(object):
    def debug(self, msg):
        pass

    def warning(self, msg):
        pass

    def error(self, msg):
        pass


class SizedTextColumn(ProgressColumn):
    """A column containing text."""

    def __init__(
        self,
        text_format: str,
        style: StyleType = "none",
        justify: JustifyMethod = "left",
        markup: bool = True,
        highlighter: Highlighter = None,
        overflow: Optional[OverflowMethod] = None,
        width: int = 20,
    ) -> None:
        self.text_format = text_format
        self.justify: JustifyMethod = justify
        self.style = style
        self.markup = markup
        self.highlighter = highlighter
        self.overflow: Optional[OverflowMethod] = overflow
        self.width = width
        super().__init__()

    def render(self, task: "Task") -> Text:
        _text = self.text_format.format(task=task)
        if self.markup:
            text = Text.from_markup(_text, style=self.style, justify=self.justify)
        else:
            text = Text(_text, style=self.style, justify=self.justify)
        if self.highlighter:
            self.highlighter.highlight(text)

        text.truncate(max_width=self.width, overflow=self.overflow, pad=True)
        return text


class DisplayManager:
    def __init__(self):

        # ! Change color system if "legacy" windows terminal to prevent wrong colors displaying
        self.is_legacy = detect_legacy_windows()

        # ! dumb_terminals automatically handled by rich. Color system is too but it is incorrect
        # ! for legacy windows ... so no color for y'all.
        self.console = Console(
            theme=custom_theme, color_system="truecolor" if not self.is_legacy else None
        )

        self._rich_progress_bar = Progress(
            SizedTextColumn(
                "[white]{task.description}",
                overflow="ellipsis",
                width=int(self.console.width / 3),
            ),
            SizedTextColumn("{task.fields[message]}", width=18, style="nonimportant"),
            BarColumn(bar_width=None, finished_style="green"),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
            console=self.console,
            # ! Normally when you exit the progress context manager (or call stop())
            # ! the last refreshed display remains in the terminal with the cursor on
            # ! the following line. You can also make the progress display disappear on
            # ! exit by setting transient=True on the Progress constructor
            transient=self.is_legacy,
        )

        self.song_count = 0
        self.overall_task_id = None
        self.overall_progress = 0
        self.overall_total = 100
        self.overall_completed_tasks = 0
        self.quiet = False

        # ! Basically a wrapper for rich's: with ... as ...
        self._rich_progress_bar.__enter__()

    def print(self, *text, color="green"):
        """
        `text` : `any`  Text to be printed to screen
        Use this self.print to replace default print().
        """

        if self.quiet:
            return

        line = ""
        for item in text:
            line += str(item) + " "

        if color:
            self._rich_progress_bar.console.print(f"[{color}]{line}")
        else:
            self._rich_progress_bar.console.print(line)

    def set_song_count_to(self, song_count: int) -> None:
        """
        `int` `song_count` : number of songs being downloaded
        RETURNS `~`
        sets the size of the progressbar based on the number of songs in the current
        download set
        """

        # ! all calculations are based of the arbitrary choice that 1 song consists of
        # ! 100 steps/points/iterations
        self.song_count = song_count

        self.overall_total = 100 * song_count

        if self.song_count > 4:
            self.overall_task_id = self._rich_progress_bar.add_task(
                description="Total",
                process_id="0",
                message=f"{self.overall_completed_tasks}/{int(self.overall_total / 100)} complete",
                total=self.overall_total,
                visible=(not self.quiet),
            )

    def update_overall(self):
        """
        Updates the overall progress bar.
        """

        # If the overall progress bar exists
        if self.overall_task_id is not None:
            self._rich_progress_bar.update(
                self.overall_task_id,
                message=f"{self.overall_completed_tasks}/{int(self.overall_total / 100)} complete",
                completed=self.overall_progress,
            )

    def new_progress_tracker(self, songObj):
        """
        returns new instance of `_ProgressTracker` that follows the `songObj` download subprocess
        """
        return _ProgressTracker(self, songObj)

    def close(self) -> None:
        """
        clean up rich
        """

        self._rich_progress_bar.stop()


# ========================
# === Progress Tracker ===
# ========================


class _ProgressTracker:
    def __init__(self, parent, song_object: SongObject):
        self.parent = parent
        self.song_object = song_object

        self.progress = 0
        self.old_progress = 0
        self.download_id = 0
        self.status = ""

        self.task_id = self.parent._rich_progress_bar.add_task(
            description=song_object.display_name,
            process_id=str(self.download_id),
            message="Download Started",
            total=100,
            completed=self.progress,
            start=False,
            visible=(not self.parent.quiet),
        )

    def notify_download_skip(self) -> None:
        """
        updates progress bar to reflect a song being skipped
        """

        self.progress = 100
        self.update("Skipping")

    def ytdl_progress_hook(self, data) -> None:
        """
        Progress hook for youtube-dl. It is called each time
        bytes are read from youtube.
        """

        # ! This will be called until download is complete, i.e we get an overall

        if data["status"] == "downloading":
            file_bytes = data["total_bytes"]
            downloaded_bytes = data["downloaded_bytes"]

            if file_bytes and downloaded_bytes:
                self.progress = downloaded_bytes / file_bytes * 90

            self.update("Downloading")

    def notify_youtube_download_completion(self) -> None:
        """
        updates progresbar to reflect a audio conversion being completed
        """

        self.progress = 90  # self.progress + 5
        self.update("Converting")

    def notify_conversion_completion(self) -> None:
        """
        updates progresbar to reflect a audio conversion being completed
        """

        self.progress = 95  # self.progress + 5
        self.update("Tagging")

    def notify_download_completion(self) -> None:
        """
        updates progresbar to reflect a download being completed
        """

        # ! Download completion implie ID# tag embedding was just finished
        self.progress = 100  # self.progress + 5
        self.update("Done")

    def notify_error(self, e, tb):
        """
        `e` : error message
        `tb` : traceback
        Freezes the progress bar and prints the traceback received
        """
        self.update(message="Error " + self.status)

        message = f"Error: {e}\tWhile {self.status}: {self.song_object.display_name}\n {str(tb)}"
        self.parent.print(message, color="red")

    def update(self, message=""):
        """
        Called at every event.
        """

        self.status = message

        # The change in progress since last update
        delta = self.progress - self.old_progress

        # Update the progress bar
        # ! `start_task` called everytime to ensure progress is remove from indeterminate state
        self.parent._rich_progress_bar.start_task(self.task_id)
        self.parent._rich_progress_bar.update(
            self.task_id,
            description=self.song_object.display_name,
            process_id=str(self.download_id),
            message=message,
            completed=self.progress,
        )

        # If task is complete
        if self.progress == 100 or message == "Error":
            self.parent.overall_completed_tasks += 1
            if self.parent.is_legacy:
                self.parent._rich_progress_bar.remove_task(self.task_id)

        # Update the overall progress bar
        self.parent.overall_progress += delta
        self.parent.update_overall()

        self.old_progress = self.progress



from argparse import ArgumentParser, RawDescriptionHelpFormatter

help_notice = r"""  # noqa: E501
To download a song run,
    spotdl [trackUrl]
    ex. spotdl https://open.spotify.com/track/0VjIjW4GlUZAMYd2vXMi3b

To download a album run,
    spotdl [albumUrl]
    ex. spotdl https://open.spotify.com/album/4yP0hdKOZPNshxUOjY0cZj

To download a playlist, run:
    spotdl [playlistUrl]
    ex. spotdl https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID

To download your saved songs, run:
    spotdl --user-auth saved

To download all songs from an artist run:
    spotdl [artistUrl]
    ex. spotdl https://open.spotify.com/artist/1fZAAHNWdSM5gqbi9o5iEA

To download youtube video with metadata from spotify run:
    spotdl "YouTubeURL|SpotifyURL"
    ex. spotdl "https://www.youtube.com/watch?v=EO7XnC1YpVo|https://open.spotify.com/track/4fzsfWzRhPawzqhX8Qt9F3"
    Note: urls that you pass have to be quoted properly ex. "YouTubeURL|SpotifyUrl"

To change output format run:
    spotdl [songUrl] --output-format mp3/m4a/flac/opus/ogg/wav
    ex. spotdl [songUrl] --output-format opus

To use ffmpeg binary that is not on PATH run:
    spotdl [songUrl] --ffmpeg path/to/your/ffmpeg.exe
    ex. spotdl [songUrl] --ffmpeg C:\ffmpeg\bin\ffmpeg.exe

To generate .m3u file for each playlist run:
    spotdl [playlistUrl] --m3u
    ex. spotdl https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID --m3u

To use youtube instead of youtube music run:
    spotdl [songUrl] --use-youtube
    ex. spotdl https://open.spotify.com/track/4fzsfWzRhPawzqhX8Qt9F3 --use-youtube

To change number of threads used when downloading songs run:
    spotdl [songUrl] --dt [number]
    ex. spotdl https://open.spotify.com/track/4fzsfWzRhPawzqhX8Qt9F3 --dt 8

To change number of threads used when searching for songs run:
    spotdl [songUrl] --st [number]
    ex. spotdl https://open.spotify.com/track/4fzsfWzRhPawzqhX8Qt9F3 --st 8

To ignore your ffmpeg version run:
    spotdl --ignore-ffmpeg-version

To search for and download a song, run, with quotation marks:
Note: This is not accurate and often causes errors.
    spotdl [songQuery]
    ex. spotdl 'The Weeknd - Blinding Lights'

To resume a failed/incomplete download, run:
    spotdl [pathToTrackingFile]
    ex. spotdl 'The Weeknd - Blinding Lights.spotdlTrackingFile'

    .spotdlTrackingFiles are automatically created when a download starts and deleted on completion

You can queue up multiple download tasks by separating the arguments with spaces:
    spotdl [songQuery1] [albumUrl] [songQuery2] ... (order does not matter)
    ex. spotdl 'The Weeknd - Blinding Lights'
            https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID ...

You can use the --debug-termination flag to figure out where in the code spotdl got stuck.

spotDL downloads up to 4 songs in parallel, so for a faster experience,
download albums and playlist, rather than tracks.
"""


def parse_arguments():
    # Initialize argument parser
    parser = ArgumentParser(
        prog="spotdl",
        description=help_notice,
        formatter_class=RawDescriptionHelpFormatter,
    )

    # Search query
    parser.add_argument(
        "query", type=str, nargs="+", help="URL/String for a song/album/playlist/artist"
    )

    # Option to enable debug termination
    parser.add_argument("--debug-termination", action="store_true")

    # Option to specify output directory
    parser.add_argument("--output", "-o", help="Output directory path")

    # Option to specify output format
    parser.add_argument(
        "--output-format",
        "--of",
        help="Output format",
        choices={"mp3", "m4a", "flac", "ogg", "opus", "wav"},
        default="mp3",
    )

    # Option to enable user auth
    parser.add_argument(
        "--user-auth", help="Use User Authentication", action="store_true"
    )

    # Option to use youtube instead of youtube music
    parser.add_argument(
        "--use-youtube", help="Use youtube instead of YTM", action="store_true"
    )

    # Option to specify path to local ffmpeg
    parser.add_argument("-f", "--ffmpeg", help="Path to ffmpeg", dest="ffmpeg")

    # Option to ignore ffmpeg version
    parser.add_argument(
        "--ignore-ffmpeg-version", help="Ignore ffmpeg version", action="store_true"
    )

    # Option to specify number of threads to use when downloading songs
    parser.add_argument(
        "--download-threads",
        "--dt",
        help="Number of threads used when downloading songs",
        type=int,
        default=4,
    )

    # Option to specify number of threads to use when searching for songs
    parser.add_argument(
        "--search-threads",
        "--st",
        help="Number of threads used when searching for songs",
        type=int,
        default=1,
    )

    # Option to generate .m3u
    parser.add_argument(
        "--generate-m3u",
        "--m3u",
        help="Generate .m3u file for each playlist",
        action="store_true",
    )

    # Option to print ffmpeg version
    parser.add_argument(
        "-v",
        "--version",
        help="Show spotDL's version and exit",
        action="store_true",
    )

    return parser.parse_args()



import sys
import shutil
import asyncio
import traceback
import concurrent.futures

from pathlib import Path
from youtube_dl import YoutubeDL
from typing import List, Optional

from spotdl.search import SongObject
from spotdl.download.progress_ui_handler import YTDLLogger
from spotdl.download import ffmpeg, set_id3_data, DisplayManager, DownloadTracker


class DownloadManager:
    def __init__(self, arguments: Optional[dict] = None):
        # start a server for objects shared across processes
        self.display_manager = DisplayManager()
        self.download_tracker = DownloadTracker()

        if arguments is None:
            arguments = {}

        arguments.setdefault("ffmpeg", "ffmpeg")
        arguments.setdefault("output_format", "mp3")
        arguments.setdefault("download_threads", 4)

        if sys.platform == "win32":
            # ! ProactorEventLoop is required on Windows to run subprocess asynchronously
            # ! it is default since Python 3.8 but has to be changed for previous versions
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)

        self.loop = asyncio.get_event_loop()
        # ! semaphore is required to limit concurrent asyncio executions
        self.semaphore = asyncio.Semaphore(arguments["download_threads"])

        # ! thread pool executor is used to run blocking (CPU-bound) code from a thread
        self.thread_executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=arguments["download_threads"]
        )

        self.arguments = arguments

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.display_manager.close()

        temp_folder = Path("spotdl-temp")

        if temp_folder.exists():
            shutil.rmtree(temp_folder)

    def download_single_song(self, song_object: SongObject) -> None:
        """
        `song_object` `song` : song to be downloaded

        RETURNS `~`

        downloads the given song
        """

        self.download_tracker.clear()
        self.download_tracker.load_song_list([song_object])

        self.display_manager.set_song_count_to(1)

        self._download_asynchronously([song_object])

    def download_multiple_songs(self, song_list: List[SongObject]) -> None:
        """
        `list<song_object>` `song_list` : list of songs to be downloaded

        RETURNS `~`

        downloads the given songs in parallel
        """

        self.download_tracker.clear()
        self.download_tracker.load_song_list(song_list)

        self.display_manager.set_song_count_to(len(song_list))

        self._download_asynchronously(song_list)

    def resume_download_from_tracking_file(self, tracking_file_path: str) -> None:
        """
        `str` `tracking_file_path` : path to a .spotdlTrackingFile

        RETURNS `~`

        downloads songs present on the .spotdlTrackingFile in parallel
        """

        self.download_tracker.clear()
        self.download_tracker.load_tracking_file(tracking_file_path)

        song_list = self.download_tracker.get_song_list()

        self.display_manager.set_song_count_to(len(song_list))

        self._download_asynchronously(song_list)

    def _download_asynchronously(self, song_obj_list):
        tasks = [self._pool_download(song) for song in song_obj_list]
        # call all task asynchronously, and wait until all are finished
        self.loop.run_until_complete(asyncio.gather(*tasks))

    async def _pool_download(self, song_obj: SongObject):
        # ! Run asynchronous task in a pool to make sure that all processes
        # ! don't run at once.

        # tasks that cannot acquire semaphore will wait here until it's free
        # only certain amount of tasks can acquire the semaphore at the same time
        async with self.semaphore:
            return await self.download_song(song_obj)

    async def download_song(self, song_object: SongObject) -> None:
        """
        `song_object` `song_object` : song to be downloaded

        RETURNS `~`

        Downloads, Converts, Normalizes song & embeds metadata as ID3 tags.
        """

        display_progress_tracker = self.display_manager.new_progress_tracker(
            song_object
        )

        # ! since most errors are expected to happen within this function, we wrap in
        # ! exception catcher to prevent blocking on multiple downloads
        try:

            # ! all YouTube downloads are to .\Temp; they are then converted and put into .\ and
            # ! finally followed up with ID3 metadata tags

            # ! we explicitly use the os.path.join function here to ensure download is
            # ! platform agnostic

            # Create a spotdl-temp folder if not present
            temp_folder = Path("spotdl-temp")

            if not temp_folder.exists():
                temp_folder.mkdir()

            converted_file_path = _get_converted_file_path(
                song_object, self.arguments["output_format"]
            )

            # if a song is already downloaded skip it
            if converted_file_path.is_file():
                if self.display_manager:
                    display_progress_tracker.notify_download_skip()
                if self.download_tracker:
                    self.download_tracker.notify_download_completion(song_object)

                # ! None is the default return value of all functions, we just explicitly define
                # ! it here as a continent way to avoid executing the rest of the function.
                return None

            if self.arguments["output_format"] == "m4a":
                ytdl_format = "bestaudio[ext=m4a]"
            elif self.arguments["output_format"] == "opus":
                ytdl_format = "bestaudio[ext=webm]"
            else:
                ytdl_format = "bestaudio"

            # download Audio from YouTube
            audio_handler = YoutubeDL(
                {
                    "format": ytdl_format,
                    "outtmpl": f"{str(temp_folder)}/%(id)s.%(ext)s",
                    "quiet": True,
                    "no_warnings": True,
                    "ignoreerrors": True,
                    "logger": YTDLLogger(),
                    "progress_hooks": [display_progress_tracker.ytdl_progress_hook]
                    if display_progress_tracker
                    else [],
                }
            )

            try:
                downloaded_file_path_string = await self._perform_audio_download_async(
                    converted_file_path.name.split(".")[0],
                    temp_folder,
                    audio_handler,
                    song_object.youtube_link,
                )
            except Exception:
                print(
                    f'Unable to get audio stream for "{song_object.song_name}" '
                    f'by "{song_object.contributing_artists[0]}" '
                    f'from video "{song_object.youtube_link}"'
                )
                return None

            if downloaded_file_path_string is None:
                return None

            if display_progress_tracker:
                display_progress_tracker.notify_youtube_download_completion()

            downloaded_file_path = Path(downloaded_file_path_string)

            if self.arguments["output_format"] != "m4a":
                ffmpeg_success = await ffmpeg.convert(
                    downloaded_file_path=downloaded_file_path,
                    converted_file_path=converted_file_path,
                    output_format=self.arguments["output_format"],
                    ffmpeg_path=self.arguments["ffmpeg"],
                )
            else:
                downloaded_file_path.rename(converted_file_path)
                ffmpeg_success = True

            if display_progress_tracker:
                display_progress_tracker.notify_conversion_completion()

            if ffmpeg_success is False:
                # delete the file that wasn't successfully converted
                converted_file_path.unlink()
            else:
                # if a file was successfully downloaded, tag it
                set_id3_data(
                    converted_file_path, song_object, self.arguments["output_format"]
                )

            # Do the necessary cleanup
            if display_progress_tracker:
                display_progress_tracker.notify_download_completion()

            if self.download_tracker:
                self.download_tracker.notify_download_completion(song_object)

            # delete the unnecessary YouTube download File
            if downloaded_file_path and downloaded_file_path.is_file():
                downloaded_file_path.unlink()

        except Exception as e:
            tb = traceback.format_exc()
            if display_progress_tracker:
                display_progress_tracker.notify_error(e, tb)
            else:
                raise e

    async def _perform_audio_download_async(
        self, converted_file_name, temp_folder, track_audio_stream, youtube_link
    ):
        # ! The following function calls blocking code, which would block whole event loop.
        # ! Therefore it has to be called in a separate thread via ThreadPoolExecutor. This
        # ! is not a problem, since GIL is released for the I/O operations, so it shouldn't
        # ! hurt performance.
        return await self.loop.run_in_executor(
            self.thread_executor,
            self._perform_audio_download,
            converted_file_name,
            temp_folder,
            track_audio_stream,
            youtube_link,
        )

    def _perform_audio_download(
        self, converted_file_name, temp_folder, audio_handler, youtube_link
    ):
        # ! The actual download, if there is any error, it'll be here,
        try:
            data = audio_handler.extract_info(youtube_link)
            downloaded_file_path = Path(temp_folder / f"{data['id']}.{data['ext']}")

            return downloaded_file_path
        except Exception as e:  # noqa:E722
            # ! This is equivalent to a failed download, we do nothing, the song remains on
            # ! download_trackers download queue and all is well...

            temp_files = Path(temp_folder).glob(f"{converted_file_name}.*")
            for temp_file in temp_files:
                temp_file.unlink()

            raise e


# ========================
# === Helper function ===
# ========================


def _sanitize_filename(input_str: str) -> str:
    output = input_str

    # ! this is windows specific (disallowed chars)
    output = "".join(char for char in output if char not in "/?\\*|<>")

    # ! double quotes (") and semi-colons (:) are also disallowed characters but we would
    # ! like to retain their equivalents, so they aren't removed in the prior loop
    output = output.replace('"', "'").replace(":", "-")

    return output


def _get_smaller_file_path(input_song: SongObject, output_format: str) -> Path:
    # Only use the first artist if the song path turns out to be too long
    smaller_name = f"{input_song.contributing_artists[0]} - {input_song.song_name}"

    smaller_name = _sanitize_filename(smaller_name)

    try:
        return Path(f"{smaller_name}.{output_format}").resolve()
    except (OSError, WindowsError):
        # Expected to happen in the rare case when the saved path is too long,
        # even with the short filename
        raise OSError("Cannot save song due to path issues.")


def _get_converted_file_path(song_obj: SongObject, output_format: str = None) -> Path:

    # ! we eliminate contributing artist names that are also in the song name, else we
    # ! would end up with things like 'Jetta, Mastubs - I'd love to change the world
    # ! (Mastubs REMIX).mp3' which is kinda an odd file name.

    # also make sure that main artist is included in artistStr even if they
    # are in the song name, for example
    # Lil Baby - Never Recover (Lil Baby & Gunna, Drake).mp3

    artists_filtered = []

    if output_format is None:
        output_format = "mp3"

    for artist in song_obj.contributing_artists:
        if artist.lower() not in song_obj.song_name:
            artists_filtered.append(artist)
        elif artist.lower() is song_obj.contributing_artists[0].lower():
            artists_filtered.append(artist)

    artist_str = ", ".join(artists_filtered)

    converted_file_name = _sanitize_filename(
        f"{artist_str} - {song_obj.song_name}.{output_format}"
    )

    converted_file_path = Path(converted_file_name)

    # ! Checks if a file name is too long (256 max on both linux and windows)
    try:
        if len(str(converted_file_path.resolve().name)) > 256:
            print("Path was too long. Using Small Path.")
            return _get_smaller_file_path(song_obj, output_format)
    except (OSError, WindowsError):
        return _get_smaller_file_path(song_obj, output_format)

    return converted_file_path



import os
import sys
import signal
import pkg_resources

from spotdl.download import ffmpeg, DownloadManager
from spotdl.parsers import parse_arguments, parse_query
from spotdl.search import SpotifyClient


def console_entry_point():
    """
    This is where all the console processing magic happens.
    Its super simple, rudimentary even but, it's dead simple & it works.
    """

    # If -v parameter is specified print version and exit
    if len(sys.argv) >= 2 and sys.argv[1] in ["-v", "--version"]:
        version = pkg_resources.require("spotdl")[0].version
        print(version)
        sys.exit(0)

    # Parser arguments
    arguments = parse_arguments()

    # Convert arguments to dict
    args_dict = vars(arguments)

    # Check if ffmpeg has correct version, if not exit
    if (
        ffmpeg.has_correct_version(
            arguments.ignore_ffmpeg_version, arguments.ffmpeg or "ffmpeg"
        )
        is False
    ):
        sys.exit(1)

    if "saved" in arguments.query and not arguments.user_auth:
        arguments.user_auth = True
        print(
            "Detected 'saved' in command line, but no --user-auth flag. Enabling Anyways."
        )
        print("Please Log In...")

    # Initialize spotify client
    SpotifyClient.init(
        client_id="5f573c9620494bae87890c0f08a60293",
        client_secret="212476d9b0f3472eaa762d90b19b0ba8",
        user_auth=arguments.user_auth,
    )

    # Change directory if user specified correct output path
    if arguments.output:
        if not os.path.isdir(arguments.output):
            sys.exit("The output directory doesn't exist.")
        print(f"Will download to: {os.path.abspath(arguments.output)}")
        os.chdir(arguments.output)

    # Start download manager
    with DownloadManager(args_dict) as downloader:
        if not arguments.debug_termination:

            def graceful_exit(signal, frame):
                downloader.display_manager.close()
                sys.exit(0)

            signal.signal(signal.SIGINT, graceful_exit)
            signal.signal(signal.SIGTERM, graceful_exit)

        # Find tracking files in queries
        tracking_files = [
            query for query in arguments.query if query.endswith(".spotdlTrackingFile")
        ]

        # Restart downloads
        for tracking_file in tracking_files:
            print("Preparing to resume download...")
            downloader.resume_download_from_tracking_file(tracking_file)

        # Get songs
        song_list = parse_query(
            arguments.query,
            arguments.output_format,
            arguments.use_youtube,
            arguments.generate_m3u,
            arguments.search_threads,
        )

        # Start downloading
        if len(song_list) > 0:
            downloader.download_multiple_songs(song_list)



# ===============
# === Imports ===
# ===============

# ! Just for static typing
from ytmusicapi import YTMusic
from unidecode import unidecode
from typing import List, Optional

from spotdl.providers.provider_utils import (
    _match_percentage,
    _parse_duration,
    _create_song_title,
)


# ! YTMusic api client
ytm_client = YTMusic()


def search_and_get_best_match(
    song_name: str,
    song_artists: List[str],
    song_album_name: str,
    song_duration: int,
    isrc: str,
) -> Optional[str]:
    """
    `str` `song_name` : name of song

    `list<str>` `song_artists` : list containing name of contributing artists

    `str` `song_album_name` : name of song's album

    `int` `song_duration` : duration of the song

    `str` `isrc` :  code for identifying sound recordings and music video recordings

    RETURNS `str` : link of the best match
    """

    # if isrc is not None then we try to find song with it
    if isrc is not None:
        isrc_results = _query_and_simplify(isrc, "songs")

        if len(isrc_results) == 1:
            isrc_result = isrc_results[0]

            name_match = isrc_result["name"].lower() == song_name.lower()

            delta = isrc_result["length"] - song_duration
            non_match_value = (delta ** 2) / song_duration * 100

            time_match = 100 - non_match_value

            if (
                isrc_result is not None
                and "link" in isrc_result
                and name_match is True
                and time_match > 90
            ):
                return isrc_result["link"]

    song_title = _create_song_title(song_name, song_artists)

    # Query YTM by songs only first, this way if we get correct result on the first try
    # we don't have to make another request to ytmusic api that could result in us
    # getting rate limited sooner
    song_results = _query_and_simplify(song_title, "songs")

    # Order results
    songs = _order_ytm_results(
        song_results, song_name, song_artists, song_album_name, song_duration
    )

    # song type results are always more accurate than video type, so if we get score of 80 or above
    # we are almost 100% sure that this is the correct link
    if len(songs) != 0:
        # get the result with highest score
        best_result = max(songs, key=lambda k: songs[k])

        if songs[best_result] >= 80:
            return best_result

    # We didn't find the correct song on the first try so now we get video type results
    # add them to song_results, and get the result with highest score
    video_results = _query_and_simplify(
        _create_song_title(song_name, song_artists), filter="videos"
    )

    # Order video results
    videos = _order_ytm_results(
        video_results, song_name, song_artists, song_album_name, song_duration
    )

    # Merge songs and video results
    results = {**songs, **videos}

    # No matches found
    if len(results) == 0:
        return None

    result_items = list(results.items())

    # Sort results by highest score
    sorted_results = sorted(result_items, key=lambda x: x[1], reverse=True)

    # ! In theory, the first 'TUPLE' in sorted_results should have the highest match
    # ! value, we send back only the link
    return sorted_results[0][0]


def _order_ytm_results(
    results: List[dict],
    song_name: str,
    song_artists: List[str],
    song_album_name: str,
    song_duration: int,
) -> dict:

    # Assign an overall avg match value to each result
    links_with_match_value = {}

    for result in results:
        # ! skip results without videoId, this happens if you are country restricted or
        # ! video is unavailabe
        if result == {}:
            continue

        # ! If there are no common words b/w the spotify and YouTube Music name, the song
        # ! is a wrong match (Like Ruelle - Madness being matched to Ruelle - Monster, it
        # ! happens without this conditional)

        # ! most song results on youtube go by $artist - $song_name, so if the spotify name
        # ! has a '-', this function would return True, a common '-' is hardly a 'common
        # ! word', so we get rid of it. Lower-caseing all the inputs is to get rid of the
        # ! troubles that arise from pythons handling of differently cased words, i.e.
        # ! 'Rhino' == 'rhino' is false though the word is same... so we lower-case both
        # ! sentences and replace any hypens(-)
        lower_song_name = song_name.lower()
        lower_result_name = result["name"].lower()

        sentence_words = lower_song_name.replace("-", " ").split(" ")

        common_word = False

        # ! check for common word
        for word in sentence_words:
            if word != "" and word in lower_result_name:
                common_word = True

        # ! if there are no common words, skip result
        if not common_word:
            continue

        # Find artist match
        # ! match  = (no of artist names in result) / (no. of artist names on spotify) * 100
        artist_match_number = 0

        # ! we use fuzzy matching because YouTube spellings might be mucked up
        if result["type"] == "song":
            for artist in song_artists:
                if _match_percentage(
                    unidecode(artist.lower()), unidecode(result["artist"]).lower(), 85
                ):
                    artist_match_number += 1
        else:
            # ! i.e if video
            for artist in song_artists:
                # ! something like _match_percentage('rionos', 'aiobahn, rionos Motivation
                # ! (remix)' would return 100, so we're absolutely corrent in matching
                # ! artists to song name.
                if _match_percentage(
                    unidecode(artist.lower()), unidecode(result["name"]).lower(), 85
                ):
                    artist_match_number += 1

            # we didn't find artist in the video title, so we fallback to
            # detecting song artist in the channel name
            # I am not sure if this won't create false positives
            if artist_match_number == 0:
                for artist in song_artists:
                    if _match_percentage(
                        unidecode(artist.lower()),
                        unidecode(result["artist"].lower()),
                        85,
                    ):
                        artist_match_number += 1

        # ! Skip if there are no artists in common, (else, results like 'Griffith Swank -
        # ! Madness' will be the top match for 'Ruelle - Madness')
        if artist_match_number == 0:
            continue

        artist_match = (artist_match_number / len(song_artists)) * 100

        song_title = _create_song_title(song_name, song_artists)

        # Find name match and drop results below 60%
        # this needs more testing
        if result["type"] == "song":
            name_match = round(
                _match_percentage(unidecode(result["name"]), unidecode(song_name), 60),
                ndigits=3,
            )
        else:
            name_match = round(
                _match_percentage(unidecode(result["name"]), unidecode(song_title), 60),
                ndigits=3,
            )

        # skip results with name match of 0, these are obviously wrong
        # but can be identified as correct later on due to other factors
        # such as time_match or artist_match
        if name_match == 0:
            continue

        # Find album match
        # ! We assign an arbitrary value of 0 for album match in case of video results
        # ! from YouTube Music
        album_match = 0.0
        album = None

        if result["type"] == "song":
            album = result.get("album")
            if album:
                album_match = _match_percentage(album, song_album_name)

        # Find duration match
        # ! time match = 100 - (delta(duration)**2 / original duration * 100)
        # ! difference in song duration (delta) is usually of the magnitude of a few
        # ! seconds, we need to amplify the delta if it is to have any meaningful impact
        # ! wen we calculate the avg match value
        delta = result["length"] - song_duration
        non_match_value = (delta ** 2) / song_duration * 100

        time_match = 100 - non_match_value

        if result["type"] == "song":
            if album is not None:
                # Don't add album_match to average_match if song_name == result and
                # result album name != song_album_name
                if (
                    _match_percentage(album.lower(), result["name"].lower()) > 95
                    and album.lower() != song_album_name.lower()
                ):
                    average_match = (artist_match + name_match + time_match) / 3
                # Add album to average_match if song_name == result album
                # and result album name == song_album_name
                else:
                    average_match = (
                        artist_match + album_match + name_match + time_match
                    ) / 4
            else:
                average_match = (artist_match + name_match + time_match) / 3
        # Don't add album_match to average_match if we don't have information about the album
        # name in the metadata
        else:
            average_match = (artist_match + name_match + time_match) / 3

        # the results along with the avg Match
        links_with_match_value[result["link"]] = average_match

    return links_with_match_value


def _map_result_to_song_data(result: dict) -> dict:
    song_data = {}
    artists = ", ".join(map(lambda a: a["name"], result["artists"]))
    video_id = result["videoId"]

    # Ignore results without video id
    if video_id is None:
        return {}

    song_data = {
        "name": result["title"],
        "type": result["resultType"],
        "artist": artists,
        "length": _parse_duration(result.get("duration", None)),
        "link": f"https://www.youtube.com/watch?v={video_id}",
        "position": 0,
    }

    album = result.get("album")
    if album:
        song_data["album"] = album["name"]

    return song_data


def _query_and_simplify(search_term: str, filter: str) -> List[dict]:
    """
    `str` `searchTerm` : the search term you would type into YTM's search bar
    `str` `filter` : Filter for item types

    RETURNS `list<dict>`

    For structure of dict, see comment at function declaration
    """

    # ! For dict structure, see end of this function (~ln 268, ln 283) and chill, this
    # ! function ain't soo big, there are plenty of comments and blank lines

    # build and POST a query to YTM
    search_results = ytm_client.search(search_term, filter=filter)

    return list(map(_map_result_to_song_data, search_results))



import asyncio
import subprocess
import sys
import re


def has_correct_version(
    skip_version_check: bool = False, ffmpeg_path: str = "ffmpeg"
) -> bool:
    try:
        process = subprocess.Popen(
            [ffmpeg_path, "-version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )
    except FileNotFoundError:
        print("FFmpeg was not found, spotDL cannot continue.", file=sys.stderr)
        return False

    output = "".join(process.communicate())

    if skip_version_check is False:
        result = re.search(r"ffmpeg version \w?(\d+\.)?(\d+)", output)

        if result is not None:
            version = result.group(0).replace("ffmpeg version ", "")

            # remove all non numeric characters from string example: n4.3
            version = re.sub(r"[a-zA-Z]", "", version)

            if float(version) < 4.2:
                print(
                    f"Your FFmpeg installation is too old ({version}), please update to 4.2+\n",
                    file=sys.stderr,
                )
                return False

            return True
        else:
            # fallback to copyright date check
            date_result = re.search(r"Copyright \(c\) \d\d\d\d\-\d\d\d\d", output)

            if date_result is not None:
                date = date_result.group(0)
                if "2021" in date or "2020" in date:
                    return True

            print("Your FFmpeg version couldn't be detected", file=sys.stderr)
            return False
    else:
        return True


async def convert(
    downloaded_file_path, converted_file_path, ffmpeg_path, output_format
) -> bool:
    # ! '-abr true' automatically determines and passes the
    # ! audio encoding bitrate to the filters and encoder. This ensures that the
    # ! sampled length of songs matches the actual length (i.e. a 5 min song won't display
    # ! as 47 seconds long in your music player, yeah that was an issue earlier.)

    downloaded_file_path = str(downloaded_file_path)
    converted_file_path = str(converted_file_path)

    formats = {
        "mp3": ["-codec:a", "libmp3lame"],
        "flac": ["-codec:a", "flac"],
        "ogg": ["-codec:a", "libvorbis"],
        "opus": ["-vn", "-c:a", "copy"],
        "m4a": ["-codec:a", "aac", "-vn"],
        "wav": [],
    }

    if output_format is None:
        output_format_command = formats["mp3"]
    else:
        output_format_command = formats[output_format]

    if ffmpeg_path is None:
        ffmpeg_path = "ffmpeg"

    arguments = [  # type: ignore
        "-i",
        downloaded_file_path,
        *output_format_command,
        "-abr",
        "true",
        "-q:a",
        "0",
        "-v",
        "debug",
        converted_file_path,
    ]

    process = await asyncio.subprocess.create_subprocess_exec(
        ffmpeg_path,
        *arguments,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    proc_out = await process.communicate()

    if proc_out[0] and proc_out[1]:
        out = str(b"".join(proc_out))
    else:
        out = ""

    if process.returncode != 0:
        message = (
            f"ffmpeg returned an error ({process.returncode})"
            f'\nffmpeg arguments: "{" ".join(arguments)}"'
            "\nffmpeg gave this output:"
            "\n=====\n"
            f"{out}"
            "\n=====\n"
        )

        print(message, file=sys.stderr)
        return False

    return True



import base64

from urllib.request import urlopen
from mutagen.oggopus import OggOpus
from mutagen.mp4 import MP4, MP4Cover
from mutagen.flac import Picture, FLAC
from mutagen.oggvorbis import OggVorbis
from mutagen.easyid3 import EasyID3, ID3
from mutagen.id3 import APIC as AlbumCover, USLT, COMM as Comment

from spotdl.search import SongObject


# Apple has specific tags - see mutagen docs -
# http://mutagen.readthedocs.io/en/latest/api/mp4.html
M4A_TAG_PRESET = {
    "album": "\xa9alb",
    "artist": "\xa9ART",
    "date": "\xa9day",
    "title": "\xa9nam",
    "year": "\xa9day",
    "originaldate": "purd",
    "comment": "\xa9cmt",
    "group": "\xa9grp",
    "writer": "\xa9wrt",
    "genre": "\xa9gen",
    "tracknumber": "trkn",
    "albumartist": "aART",
    "discnumber": "disk",
    "cpil": "cpil",
    "albumart": "covr",
    "copyright": "cprt",
    "tempo": "tmpo",
    "lyrics": "\xa9lyr",
    "explicit": "rtng",
}

TAG_PRESET = {}
for key in M4A_TAG_PRESET.keys():
    TAG_PRESET[key] = key


def _set_id3_mp3(converted_file_path: str, song_object: SongObject):
    # embed song details
    # ! we save tags as both ID3 v2.3 and v2.4
    # ! The simple ID3 tags
    audio_file = EasyID3(converted_file_path)

    audio_file = _embed_mp3_metadata(audio_file, song_object)

    # ! save as both ID3 v2.3 & v2.4 as v2.3 isn't fully features and
    # ! windows doesn't support v2.4 until later versions of Win10
    audio_file.save(v2_version=3)

    audio_file = _embed_mp3_cover(audio_file, song_object, converted_file_path)
    audio_file = _embed_mp3_lyrics(audio_file, song_object)
    # ! setting song links as comment (helpful for devs)
    audio_file.add(Comment(encoding=3, text=song_object.youtube_link))

    audio_file.save(v2_version=3)


def _set_id3_m4a(converted_file_path: str, song_object: SongObject):
    # embed song details
    # ! we save tags as both ID3 v2.3 and v2.4
    # ! The simple ID3 tags
    audio_file = MP4(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_object, "m4a", M4A_TAG_PRESET)
    audio_file = _embed_m4a_metadata(audio_file, song_object)

    audio_file.save()


def _set_id3_flac(converted_file_path: str, song_object: SongObject):
    audio_file = FLAC(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_object, "flac")
    audio_file = _embed_ogg_metadata(audio_file, song_object)
    audio_file = _embed_cover(audio_file, song_object, "flac")

    audio_file.save()


def _set_id3_opus(converted_file_path: str, song_object: SongObject):
    audio_file = OggOpus(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_object, "opus")
    audio_file = _embed_ogg_metadata(audio_file, song_object)
    audio_file = _embed_cover(audio_file, song_object, "opus")

    audio_file.save()


def _set_id3_ogg(converted_file_path: str, song_object: SongObject):
    audio_file = OggVorbis(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_object, "ogg")
    audio_file = _embed_ogg_metadata(audio_file, song_object)
    audio_file = _embed_cover(audio_file, song_object, "ogg")

    audio_file.save()


def _embed_mp3_metadata(audio_file, song_object: SongObject):

    # ! Get rid of all existing ID3 tags (if any exist)
    audio_file.delete()

    # ! song name
    audio_file["title"] = song_object.song_name
    audio_file["titlesort"] = song_object.song_name

    # ! track number
    audio_file["tracknumber"] = str(song_object.track_number)

    # ! disc number
    audio_file["discnumber"] = str(song_object.disc_number)

    # ! genres (pretty pointless if you ask me)
    # ! we only apply the first available genre as ID3 v2.3 doesn't support multiple
    # ! genres and ~80% of the world PC's run Windows - an OS with no ID3 v2.4 support
    genres = song_object.genres
    if len(genres) > 0:
        audio_file["genre"] = genres[0]

    # ! all involved artists
    audio_file["artist"] = song_object.contributing_artists

    # ! album name
    audio_file["album"] = song_object.album_name

    # ! album artist (all of 'em)
    audio_file["albumartist"] = song_object.album_artists

    # ! album release date (to what ever precision available)
    audio_file["date"] = song_object.album_release
    audio_file["originaldate"] = song_object.album_release

    return audio_file


def _embed_mp3_cover(audio_file, song_object, converted_file_path):
    # ! setting the album art
    audio_file = ID3(converted_file_path)
    rawAlbumArt = urlopen(song_object.album_cover_url).read()
    audio_file["APIC"] = AlbumCover(
        encoding=3, mime="image/jpeg", type=3, desc="Cover", data=rawAlbumArt
    )

    return audio_file


def _embed_mp3_lyrics(audio_file, song_object):
    # ! setting the lyrics
    lyrics = song_object.lyrics
    USLTOutput = USLT(encoding=3, lang=u"eng", desc=u"desc", text=lyrics)
    audio_file["USLT::'eng'"] = USLTOutput

    return audio_file


def _embed_m4a_metadata(audio_file, song_object: SongObject):
    # set year
    years = song_object.album_release.split("-")
    if len(years) >= 1:
        audio_file[M4A_TAG_PRESET["year"]] = years[0]

    # set youtube link as comment
    youtube_link = song_object.youtube_link
    if youtube_link:
        audio_file[M4A_TAG_PRESET["comment"]] = youtube_link

    # set lyrics
    lyrics = song_object.lyrics
    if lyrics:
        audio_file[M4A_TAG_PRESET["lyrics"]] = lyrics

    # Explicit values: Dirty: 4, Clean: 2, None: 0
    audio_file[M4A_TAG_PRESET["explicit"]] = (0,)
    try:
        audio_file[M4A_TAG_PRESET["albumart"]] = [
            MP4Cover(
                urlopen(song_object.album_cover_url).read(),
                imageformat=MP4Cover.FORMAT_JPEG,
            )
        ]
    except IndexError:
        pass

    return audio_file


def _embed_basic_metadata(
    audio_file, song_object: SongObject, encoding, preset=TAG_PRESET
):

    # set main artist
    artists = song_object.contributing_artists
    if len(artists):
        audio_file[preset["artist"]] = artists[0]
        audio_file[preset["albumartist"]] = artists[0]

    # set song title
    song_title = song_object.song_name
    if song_title:
        audio_file[preset["title"]] = song_title

    # set album name
    album_name = song_object.album_name
    if album_name:
        audio_file[preset["album"]] = album_name

    # set release data
    release_data = song_object.album_release
    if release_data:
        audio_file[preset["date"]] = release_data
        audio_file[preset["originaldate"]] = release_data

    # set genre
    genres = song_object.genres
    if len(genres) >= 1:
        audio_file[preset["genre"]] = genres[0]

    # set disc number
    disc_number = song_object.disc_number
    if disc_number:
        if encoding in ["flac", "ogg", "opus"]:
            audio_file[preset["discnumber"]] = str(disc_number)
        else:
            audio_file[preset["discnumber"]] = [(disc_number, 0)]

    # set track number
    track_number = song_object.track_number
    if track_number:
        if encoding in ["flac", "ogg", "opus"]:
            audio_file[preset["tracknumber"]] = str(track_number)
        else:
            audio_file[preset["tracknumber"]] = [(track_number, 0)]

    return audio_file


def _embed_ogg_metadata(audio_file, song_object: SongObject):
    # set year
    years = song_object.album_release.split("-")
    if len(years) >= 1:
        audio_file["year"] = years[0]

    # set youtube link as comment
    youtube_link = song_object.youtube_link
    if youtube_link:
        audio_file["comment"] = youtube_link

    # set lyrics
    lyrics = song_object.lyrics
    if lyrics:
        audio_file["lyrics"] = lyrics

    return audio_file


def _embed_cover(audio_file, song_object, encoding):
    image = Picture()
    image.type = 3
    image.desc = "Cover"
    image.mime = "image/jpeg"
    image.data = urlopen(song_object.album_cover_url).read()

    if encoding == "flac":
        audio_file.add_picture(image)
    elif encoding == "ogg" or encoding == "opus":
        # From the Mutagen docs (https://mutagen.readthedocs.io/en/latest/user/vcomment.html)
        image_data = image.write()
        encoded_data = base64.b64encode(image_data)
        vcomment_value = encoded_data.decode("ascii")
        audio_file["metadata_block_picture"] = [vcomment_value]

    return audio_file


SET_ID3_FUNCTIONS = {
    "mp3": _set_id3_mp3,
    "flac": _set_id3_flac,
    "opus": _set_id3_opus,
    "ogg": _set_id3_ogg,
    "m4a": _set_id3_m4a,
}


def set_id3_data(converted_file_path, song_object, output_format):
    function = SET_ID3_FUNCTIONS.get(output_format)
    if function:
        function(converted_file_path, song_object)



back_in_black = [
    {
        "youtube_link": "https://www.youtube.com/watch?v=9vWNauaZAgg",
        "raw_track_meta": {
            "album": {
                "album_type": "album",
                "artists": [
                    {
                        "external_urls": {
                            "spotify": "https://open.spotify.com/artist/711MCceyCBcFnzjGY4Q7Un"
                        },
                        "href": "https://api.spotify.com/v1/artists/711MCceyCBcFnzjGY4Q7Un",
                        "id": "711MCceyCBcFnzjGY4Q7Un",
                        "name": "AC/DC",
                        "type": "artist",
                        "uri": "spotify:artist:711MCceyCBcFnzjGY4Q7Un",
                    }
                ],
                "available_markets": [
                    "AD",
                    "AE",
                    "AL",
                    "AR",
                    "AT",
                    "AU",
                    "BA",
                    "BE",
                    "BG",
                    "BH",
                    "BO",
                    "BR",
                    "BY",
                    "CA",
                    "CH",
                    "CL",
                    "CO",
                    "CR",
                    "CY",
                    "CZ",
                    "DE",
                    "DK",
                    "DO",
                    "DZ",
                    "EC",
                    "EE",
                    "EG",
                    "ES",
                    "FI",
                    "FR",
                    "GB",
                    "GR",
                    "GT",
                    "HK",
                    "HN",
                    "HR",
                    "HU",
                    "ID",
                    "IE",
                    "IL",
                    "IN",
                    "IS",
                    "IT",
                    "JO",
                    "JP",
                    "KW",
                    "KZ",
                    "LB",
                    "LI",
                    "LT",
                    "LU",
                    "LV",
                    "MA",
                    "MC",
                    "MD",
                    "ME",
                    "MK",
                    "MT",
                    "MX",
                    "MY",
                    "NI",
                    "NL",
                    "NO",
                    "NZ",
                    "OM",
                    "PA",
                    "PE",
                    "PH",
                    "PL",
                    "PS",
                    "PT",
                    "PY",
                    "QA",
                    "RO",
                    "RS",
                    "RU",
                    "SA",
                    "SE",
                    "SG",
                    "SI",
                    "SK",
                    "SV",
                    "TH",
                    "TN",
                    "TR",
                    "TW",
                    "UA",
                    "US",
                    "UY",
                    "VN",
                    "XK",
                    "ZA",
                ],
                "external_urls": {
                    "spotify": "https://open.spotify.com/album/6mUdeDZCsExyJLMdAfDuwh"
                },
                "href": "https://api.spotify.com/v1/albums/6mUdeDZCsExyJLMdAfDuwh",
                "id": "6mUdeDZCsExyJLMdAfDuwh",
                "images": [
                    {
                        "height": 640,
                        "url": "https://i.scdn.co/image/ab67616d0000b2730b51f8d91f3a21e8426361ae",
                        "width": 640,
                    },
                    {
                        "height": 300,
                        "url": "https://i.scdn.co/image/ab67616d00001e020b51f8d91f3a21e8426361ae",
                        "width": 300,
                    },
                    {
                        "height": 64,
                        "url": "https://i.scdn.co/image/ab67616d000048510b51f8d91f3a21e8426361ae",
                        "width": 64,
                    },
                ],
                "name": "Back In Black",
                "release_date": "1980-07-25",
                "release_date_precision": "day",
                "total_tracks": 10,
                "type": "album",
                "uri": "spotify:album:6mUdeDZCsExyJLMdAfDuwh",
            },
            "artists": [
                {
                    "external_urls": {
                        "spotify": "https://open.spotify.com/artist/711MCceyCBcFnzjGY4Q7Un"
                    },
                    "href": "https://api.spotify.com/v1/artists/711MCceyCBcFnzjGY4Q7Un",
                    "id": "711MCceyCBcFnzjGY4Q7Un",
                    "name": "AC/DC",
                    "type": "artist",
                    "uri": "spotify:artist:711MCceyCBcFnzjGY4Q7Un",
                }
            ],
            "available_markets": [
                "AD",
                "AE",
                "AL",
                "AR",
                "AT",
                "AU",
                "BA",
                "BE",
                "BG",
                "BH",
                "BO",
                "BR",
                "BY",
                "CA",
                "CH",
                "CL",
                "CO",
                "CR",
                "CY",
                "CZ",
                "DE",
                "DK",
                "DO",
                "DZ",
                "EC",
                "EE",
                "EG",
                "ES",
                "FI",
                "FR",
                "GB",
                "GR",
                "GT",
                "HK",
                "HN",
                "HR",
                "HU",
                "ID",
                "IE",
                "IL",
                "IN",
                "IS",
                "IT",
                "JO",
                "JP",
                "KW",
                "KZ",
                "LB",
                "LI",
                "LT",
                "LU",
                "LV",
                "MA",
                "MC",
                "MD",
                "ME",
                "MK",
                "MT",
                "MX",
                "MY",
                "NI",
                "NL",
                "NO",
                "NZ",
                "OM",
                "PA",
                "PE",
                "PH",
                "PL",
                "PS",
                "PT",
                "PY",
                "QA",
                "RO",
                "RS",
                "RU",
                "SA",
                "SE",
                "SG",
                "SI",
                "SK",
                "SV",
                "TH",
                "TN",
                "TR",
                "TW",
                "UA",
                "US",
                "UY",
                "VN",
                "XK",
                "ZA",
            ],
            "disc_number": 1,
            "duration_ms": 255493,
            "explicit": False,
            "external_ids": {"isrc": "AUAP08000046"},
            "external_urls": {
                "spotify": "https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj"
            },
            "href": "https://api.spotify.com/v1/tracks/08mG3Y1vljYA6bvDt4Wqkj",
            "id": "08mG3Y1vljYA6bvDt4Wqkj",
            "is_local": False,
            "name": "Back In Black",
            "popularity": 85,
            "preview_url": "https://p.scdn.co/mp3-preview/506bc9a0485990973449b0467791e06174371284?cid=4fe3fecfe5334023a1472516cc99d805",
            "track_number": 6,
            "type": "track",
            "uri": "spotify:track:08mG3Y1vljYA6bvDt4Wqkj",
        },
        "raw_album_meta": {
            "external_urls": {
                "spotify": "https://open.spotify.com/artist/711MCceyCBcFnzjGY4Q7Un"
            },
            "followers": {"href": None, "total": 16821071},
            "genres": ["album rock", "australian rock", "hard rock", "rock"],
            "href": "https://api.spotify.com/v1/artists/711MCceyCBcFnzjGY4Q7Un",
            "id": "711MCceyCBcFnzjGY4Q7Un",
            "images": [
                {
                    "height": 640,
                    "url": "https://i.scdn.co/image/52f9c4f83d8642ed80a0bad0317ab940e034de08",
                    "width": 640,
                },
                {
                    "height": 320,
                    "url": "https://i.scdn.co/image/64d25927322389160df14e1966849040954b29bb",
                    "width": 320,
                },
                {
                    "height": 160,
                    "url": "https://i.scdn.co/image/a7fa12b3b307ac3f172cf1760f385b44bc29e083",
                    "width": 160,
                },
            ],
            "name": "AC/DC",
            "popularity": 88,
            "type": "artist",
            "uri": "spotify:artist:711MCceyCBcFnzjGY4Q7Un",
        },
        "raw_artist_meta": {
            "external_urls": {
                "spotify": "https://open.spotify.com/artist/711MCceyCBcFnzjGY4Q7Un"
            },
            "followers": {"href": None, "total": 16821071},
            "genres": ["album rock", "australian rock", "hard rock", "rock"],
            "href": "https://api.spotify.com/v1/artists/711MCceyCBcFnzjGY4Q7Un",
            "id": "711MCceyCBcFnzjGY4Q7Un",
            "images": [
                {
                    "height": 640,
                    "url": "https://i.scdn.co/image/52f9c4f83d8642ed80a0bad0317ab940e034de08",
                    "width": 640,
                },
                {
                    "height": 320,
                    "url": "https://i.scdn.co/image/64d25927322389160df14e1966849040954b29bb",
                    "width": 320,
                },
                {
                    "height": 160,
                    "url": "https://i.scdn.co/image/a7fa12b3b307ac3f172cf1760f385b44bc29e083",
                    "width": 160,
                },
            ],
            "name": "AC/DC",
            "popularity": 88,
            "type": "artist",
            "uri": "spotify:artist:711MCceyCBcFnzjGY4Q7Un",
        },
        "lyrics": "[Verse 1]\nBack in black, I hit the sack\nI've been too long, I'm glad to be back\nYes, I'm let loose from the noose\nThat's kept me hanging about\nI'm just looking at the sky 'cause it's getting me high\nForget the hearse, 'cause I'll never die\nI got nine lives, cat's eyes\nAbusing every one of them and running wild\n\n[Chorus]\n'Cause I'm back, yes I'm back\nWell, I'm back, yes I'm back\nWell, I'm back, back\nWell, I'm back in black\nYes, I'm back in black\n\n[Verse 2]\nBack in the back of a Cadillac\nNumber one with a bullet, I'm a power pack\nYes, I'm in a bang with a gang\nThey've got to catch me if they want me to hang\n'Cause I'm back on the track and I'm beatin' the flack\nNobody's gonna get me on another rap\nSo, look at me now, I'm just making my play\nDon't try to push your luck, just get out of my way\n\n[Chorus]\n'Cause I'm back, yes I'm back\nWell, I'm back, yes I'm back\nWell, I'm back, back\nWell, I'm back in black\nYes, I'm back in black\n\n[Guitar Solo]\n\n[Chorus]\nWell, I'm back, yes I'm back\nWell, I'm back, yes I'm back\nWell, I'm back, back\nWell, I'm back in black\nYes, I'm back in black, yow\n\n[Bridge]\nAh, yeah\nOh, yeah\nTake my love\nYeah, yeah\nYeah\nAh, hey yeah\nOoh, yeah\n\n[Chorus]\nWell, I'm back (I'm back)\nBack, well I'm (I'm back)\nBack (I'm back)\nBack (I'm back)\nBack (I'm back)\nBack, back in black\nYes, I'm back in black\n\n[Outro]\nI've hit the sack\n\n[Guitar Solo]",
    }
]



import sys

from spotdl.__main__ import console_entry_point
from spotdl.download import ffmpeg

SONGS = {
    "https://open.spotify.com/track/6CN3e26iQSj1N5lomh0mfO": "Eminem - Like Toy Soldiers.mp3",
    "https://open.spotify.com/track/3bNv3VuUOKgrf5hu3YcuRo": "Adele - Someone Like You.mp3"
}


def test_regressions(monkeypatch, tmpdir):
    """
    Download songs that caused problems in the past, to make sure they won't happen again.
    """
    monkeypatch.chdir(tmpdir)
    monkeypatch.setattr(sys, "argv", ["dummy", *SONGS.keys()])
    monkeypatch.setattr(ffmpeg, "has_correct_version", lambda *_: True)

    console_entry_point()

    assert sorted(
        [
            file.basename
            for file in tmpdir.listdir()
            if file.isfile() and file.basename.startswith(".") is False
        ]
    ) == sorted([*SONGS.values()])



from typing import List


class SongObject:

    # Constructor
    def __init__(
        self, raw_track_meta, raw_album_meta, raw_artist_meta, youtube_link, lyrics
    ):
        self._raw_track_meta = raw_track_meta
        self._raw_album_meta = raw_album_meta
        self._raw_artist_meta = raw_artist_meta
        self._youtube_link = youtube_link
        self._lyrics = lyrics

    # Equals method
    # for example song_obj1 == song_obj2
    def __eq__(self, compared_song) -> bool:
        if compared_song.data_dump == self.data_dump:
            return True
        else:
            return False

    # ================================
    # === Interface Implementation ===
    # ================================

    @property
    def youtube_link(self) -> str:
        """
        returns youtube link
        """
        return self._youtube_link

    @property
    def song_name(self) -> str:
        """
        returns songs's name.
        """

        return self._raw_track_meta["name"]

    @property
    def track_number(self) -> int:
        """
        returns song's track number from album (as in weather its the first
        or second or third or fifth track in the album)
        """

        return self._raw_track_meta["track_number"]

    @property
    def genres(self) -> List[str]:
        """
        returns a list of possible genres for the given song, the first member
        of the list is the most likely genre. returns None if genre data could
        not be found.
        """

        return self._raw_album_meta["genres"] + self._raw_artist_meta["genres"]

    @property
    def duration(self) -> float:
        """
        returns duration of song in seconds.
        """

        return round(self._raw_track_meta["duration_ms"] / 1000, ndigits=3)

    @property
    def contributing_artists(self) -> List[str]:
        """
        returns a list of all artists who worked on the song.
        The first member of the list is likely the main artist.
        """

        # we get rid of artist name that are in the song title so
        # naming the song would be as easy as
        # $contributingArtists + songName.mp3, we would want to end up with
        # 'Jetta, Mastubs - I'd love to change the world (Mastubs remix).mp3'
        # as a song name, it's dumb.

        contributingArtists = []

        for artist in self._raw_track_meta["artists"]:
            contributingArtists.append(artist["name"])

        return contributingArtists

    @property
    def disc_number(self) -> int:
        return self._raw_track_meta["disc_number"]

    @property
    def lyrics(self):
        """
        returns the lyrics of the song if found on Genius
        """

        return self._lyrics

    @property
    def display_name(self) -> str:
        """
        returns songs's display name.
        """

        return str(", ".join(self.contributing_artists) + " - " + self.song_name)

    @property
    def album_name(self) -> str:
        """
        returns name of the album that the song belongs to.
        """

        return self._raw_track_meta["album"]["name"]

    @property
    def album_artists(self) -> List[str]:
        """
        returns list of all artists who worked on the album that
        the song belongs to. The first member of the list is likely the main
        artist.
        """

        albumArtists = []

        for artist in self._raw_track_meta["album"]["artists"]:
            albumArtists.append(artist["name"])

        return albumArtists

    @property
    def album_release(self) -> str:
        """
        returns date/year of album release depending on what data is available.
        """

        return self._raw_track_meta["album"]["release_date"]

    # ! Utilities for genuine use and also for metadata freaks:

    @property
    def album_cover_url(self) -> str:
        """
        returns url of the biggest album art image available.
        """

        return self._raw_track_meta["album"]["images"][0]["url"]

    @property
    def data_dump(self) -> dict:
        """
        returns a dictionary containing the spotify-api responses as-is. The
        dictionary keys are as follows:
            - rawTrackMeta      spotify-api track details
            - rawAlbumMeta      spotify-api song's album details
            - rawArtistMeta     spotify-api song's artist details

        Avoid using this function, it is implemented here only for those super
        rare occasions where there is a need to look up other details. Why
        have to look it up seperately when it's already been looked up once?
        """

        # ! internally the only reason this exists is that it helps in saving to disk

        return {
            "youtube_link": self._youtube_link,
            "raw_track_meta": self._raw_track_meta,
            "raw_album_meta": self._raw_album_meta,
            "raw_artist_meta": self._raw_artist_meta,
            "lyrics": self._lyrics,
        }

    @property
    def file_name(self) -> str:
        return self.create_file_name(
            song_name=self._raw_track_meta["name"],
            song_artists=[artist["name"] for artist in self._raw_track_meta["artists"]],
        )

    @staticmethod
    def create_file_name(song_name: str, song_artists: List[str]) -> str:
        # build file name of converted file
        # the main artist is always included
        artistStr = song_artists[0]

        # ! we eliminate contributing artist names that are also in the song name, else we
        # ! would end up with things like 'Jetta, Mastubs - I'd love to change the world
        # ! (Mastubs REMIX).mp3' which is kinda an odd file name.
        for artist in song_artists[1:]:
            if artist.lower() not in song_name.lower():
                artistStr += ", " + artist

        convertedFileName = artistStr + " - " + song_name

        # ! this is windows specific (disallowed chars)
        convertedFileName = "".join(
            char for char in convertedFileName if char not in "/?\\*|<>"
        )

        # ! double quotes (") and semi-colons (:) are also disallowed characters but we would
        # ! like to retain their equivalents, so they aren't removed in the prior loop
        convertedFileName = convertedFileName.replace('"', "'").replace(":", "-")

        return convertedFileName



from spotdl.search import SpotifyClient


def from_url(spotify_url: str):
    if not ("open.spotify.com" in spotify_url and "track" in spotify_url):
        raise Exception(f"passed URL is not that of a track: {spotify_url}")

    # query spotify for song, artist, album details
    spotify_client = SpotifyClient()

    raw_track_meta = spotify_client.track(spotify_url)

    if raw_track_meta is None:
        raise Exception(
            "Couldn't get metadata, check if you have passed correct track id"
        )

    primary_artist_id = raw_track_meta["artists"][0]["id"]
    raw_artist_meta = spotify_client.artist(primary_artist_id)

    albumId = raw_track_meta["album"]["id"]
    raw_album_meta = spotify_client.album(albumId)

    return raw_track_meta, raw_artist_meta, raw_album_meta



from spotipy import Spotify
from spotipy.cache_handler import CacheFileHandler
from spotipy.oauth2 import SpotifyClientCredentials, SpotifyOAuth


class Singleton(type):
    """
    Singleton metaclass for SpotifyClient. Ensures that SpotifyClient is not
    instantiated without prior initialization. Every other instantiation of
    SpotifyClient will return the same instance.
    """

    _instance = None

    def __call__(cls):
        if cls._instance is None:
            raise Exception(
                "Spotify client not created. Call SpotifyClient.init"
                "(client_id, client_secret, user_auth) first."
            )
        return cls._instance

    def init(cls, client_id: str, client_secret: str, user_auth: bool) -> "Singleton":
        """
        `str` `client_id` : client id from your spotify account

        `str` `client_secret` : client secret for your client id

        `bool` `user_auth` : Determines if the Authorization Code Flow or
                   the Client Credentials Flow is used

        creates and caches a spotify client if a client doesn't exist. Can only be called
        once, multiple calls will cause an Exception.
        """

        # check if initialization has been completed, if yes, raise an Exception
        if isinstance(cls._instance, cls):
            raise Exception("A spotify client has already been initialized")

        credential_manager = None
        cache_handler = CacheFileHandler(cache_path=".spotdl-cache")

        # Use SpotifyOAuth as auth manager
        if user_auth:
            credential_manager = SpotifyOAuth(
                client_id=client_id,
                client_secret=client_secret,
                redirect_uri="http://127.0.0.1:8080/",
                scope="user-library-read",
                cache_handler=cache_handler,
            )
        # Use SpotifyClientCredentials as auth manager
        else:
            credential_manager = SpotifyClientCredentials(
                client_id=client_id,
                client_secret=client_secret,
                cache_handler=cache_handler,
            )

        # Create instance
        cls._instance = super().__call__(auth_manager=credential_manager)

        # Return instance
        return cls._instance


class SpotifyClient(Spotify, metaclass=Singleton):
    """
    This is the Spotify client meant to be used in the app. Has to be initialized first by
    calling `SpotifyClient.init(client_id, client_secret, user_auth)`.
    """

    _initialized = False

    def __init__(self, user_auth=False, *args, **kwargs):
        if user_auth:
            self._user_auth = True
        super().__init__(*args, **kwargs)
        self._initialized = True



import concurrent.futures

from pathlib import Path
from typing import Dict, List

from spotdl.providers import (
    metadata_provider,
    yt_provider,
    ytm_provider,
    provider_utils,
)
from spotdl.search import SongObject, SpotifyClient


def from_spotify_url(
    spotify_url: str, output_format: str = None, use_youtube: bool = False
) -> SongObject:
    """
    Creates song object using spotfy url

    `str` `spotify_url` : spotify url used to create song object
    `str` `output_format` : output format of the song

    returns a `SongObject`
    """

    # Set default ouput format to mp3
    if output_format is None:
        output_format = "mp3"

    # Get the Song Metadata
    print(f"Gathering Spotify Metadata for: {spotify_url}")
    raw_track_meta, raw_artist_meta, raw_album_meta = metadata_provider.from_url(
        spotify_url
    )

    if raw_track_meta is None:
        raise ValueError("Couldn't get metadata from url")

    song_name = raw_track_meta["name"]
    album_name = raw_track_meta["album"]["name"]
    isrc = raw_track_meta.get("external_ids", {}).get("isrc")
    duration = round(raw_track_meta["duration_ms"] / 1000, ndigits=3)
    contributing_artists = [artist["name"] for artist in raw_track_meta["artists"]]
    display_name = ", ".join(contributing_artists) + " - " + song_name

    # Create file name
    converted_file_name = SongObject.create_file_name(
        song_name, [artist["name"] for artist in raw_track_meta["artists"]]
    )

    # If song name is too long use only first artist
    if len(converted_file_name) > 250:
        converted_file_name = SongObject.create_file_name(
            song_name, [raw_track_meta["artists"][0]["name"]]
        )

    converted_file_path = Path(".", f"{converted_file_name}.{output_format}")

    # if a song is already downloaded skip it
    if converted_file_path.is_file():
        print(f'Skipping "{converted_file_name}" as it\'s already downloaded')
        raise OSError(f"{converted_file_name} already downloaded")

    # Get the song's downloadable audio link
    if use_youtube is True:
        print(f'Searching YouTube for "{display_name}"', end="\r")
        youtube_link = yt_provider.search_and_get_best_match(
            song_name, contributing_artists, duration, isrc
        )
    else:
        print(f'Searching YouTube Music for "{display_name}"', end="\r")
        youtube_link = ytm_provider.search_and_get_best_match(
            song_name, contributing_artists, album_name, duration, isrc
        )

    # Check if we found youtube url
    if youtube_link is None:
        print("Could not match any of the results on YouTube. Skipping")
        raise LookupError("Could not match any of the results on YouTube for")
    else:
        print(" " * (len(display_name) + 25), end="\r")
        print(f'Found YouTube URL for "{display_name}" : {youtube_link}')

    # (try to) Get lyrics from Genius
    lyrics = provider_utils._get_song_lyrics(song_name, contributing_artists)

    return SongObject(
        raw_track_meta, raw_album_meta, raw_artist_meta, youtube_link, lyrics
    )


def from_search_term(
    query: str, output_format: str = None, use_youtube: bool = False
) -> List[SongObject]:
    """
    Queries Spotify for a song and returns the best match

    `str` `query` : what you'd type into Spotify's search box
    `str` `output_format` : output format of the song

    returns a `list<SongObject>` containing Url's of each track in the given album
    """

    # get a spotify client
    spotify_client = SpotifyClient()

    # get possible matches from spotify
    result = spotify_client.search(query, type="track")

    # return first result link or if no matches are found, raise Exception
    if result is None or len(result.get("tracks", {}).get("items", [])) == 0:
        raise Exception("No song matches found on Spotify")
    else:
        song_url = "http://open.spotify.com/track/" + result["tracks"]["items"][0]["id"]
        try:
            song = from_spotify_url(song_url, output_format, use_youtube)
            return [song] if song.youtube_link is not None else []
        except (LookupError, OSError, ValueError):
            return []


def from_album(
    album_url: str,
    output_format: str = None,
    use_youtube: bool = False,
    threads: int = 1,
) -> List[SongObject]:
    """
    Create and return list containing SongObject for every song in the album

    `str` `album_url` : Spotify Url of the album whose tracks are to be retrieved
    `str` `output_format` : output format of the song

    returns a `list<SongObject>` containing Url's of each track in the given album
    """

    spotify_client = SpotifyClient()
    tracks = []

    album_response = spotify_client.album_tracks(album_url)
    if album_response is None:
        raise ValueError("Wrong album id")

    album_tracks = album_response["items"]

    # Get all tracks from album
    while album_response["next"]:
        album_response = spotify_client.next(album_response)

        # Failed to get response, break the loop
        if album_response is None:
            break

        album_tracks.extend(album_response["items"])

    # Remove songs  without id
    album_tracks = [
        track
        for track in album_tracks
        if track is not None and track.get("id") is not None
    ]

    def get_tracks(track):
        try:
            return from_spotify_url(
                "https://open.spotify.com/track/" + track["id"],
                output_format,
                use_youtube,
            )
        except (LookupError, OSError, ValueError):
            return None

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        results = executor.map(get_tracks, album_tracks)

    for song in results:
        if song is not None and song.youtube_link is not None:
            tracks.append(song)

    return tracks


def from_playlist(
    playlist_url: str,
    output_format: str = None,
    use_youtube: bool = False,
    generate_m3u: bool = False,
    threads: int = 1,
) -> List[SongObject]:
    """
    Create and return list containing SongObject for every song in the playlist

    `str` `album_url` : Spotify Url of the album whose tracks are to be retrieved
    `str` `output_format` : output format of the song

    returns a `list<SongObject>` containing Url's of each track in the given album
    """

    spotify_client = SpotifyClient()
    tracks = []

    playlist_response = spotify_client.playlist_tracks(playlist_url)
    if playlist_response is None:
        raise ValueError("Wrong playlist id")

    playlist_tracks = [
        track
        for track in playlist_response["items"]
        # check if track has id
        if track is not None
        and track.get("track") is not None
        and track["track"].get("id") is not None
    ]

    # Get all tracks from playlist
    while playlist_response["next"]:
        playlist_response = spotify_client.next(playlist_response)

        # Failed to get response, break the loop
        if playlist_response is None:
            break

        playlist_tracks.extend(playlist_response["items"])

    # Remove songs  without id
    playlist_tracks = [
        track
        for track in playlist_tracks
        if track is not None
        and track.get("track") is not None
        and track["track"].get("id") is not None
    ]

    def get_song(track):
        try:
            song = from_spotify_url(
                "https://open.spotify.com/track/" + track["track"]["id"],
                output_format,
                use_youtube,
            )
        except (LookupError, ValueError):
            return None, None
        except OSError:
            if generate_m3u:
                file_path = (
                    str(
                        provider_utils._create_song_title(
                            track["track"]["name"],
                            [artist["name"] for artist in track["track"]["artists"]],
                        )
                    )
                    + "."
                    + output_format
                    if output_format is not None
                    else "mp3"
                )
                if len(file_path) > 256:
                    file_path = (
                        str(
                            provider_utils._create_song_title(
                                track.song_name, [track.contributing_artists[0]]
                            )
                        )
                        + "."
                        + output_format
                        if output_format is not None
                        else "mp3"
                    )

                return None, f"{file_path}\n"

            return None, None
        else:
            return song, None

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        results = executor.map(get_song, playlist_tracks)

    playlist_text = ""
    for result in results:
        if result[1] is not None:
            playlist_text += result[1]

        if result[0] is not None and result[0].youtube_link is not None:
            tracks.append(result[0])

    if playlist_response and generate_m3u is True:
        playlist_data = spotify_client.playlist(playlist_url)

        if playlist_data is not None:
            playlist_name = playlist_data["name"]
        else:
            playlist_name = playlist_tracks[0]["track"]["name"]

        playlist_file = Path(f"{playlist_name}.m3u")

        with open(playlist_file, "w", encoding="utf-8") as file:
            file.write(playlist_text)

    return tracks


def from_artist(
    artist_url: str,
    output_format: str = None,
    use_youtube: bool = False,
    threads: int = 1,
) -> List[SongObject]:
    """
    `str` `artist_url` : Spotify Url of the artist whose tracks are to be
    retrieved
    returns a `list<songObj>` containing Url's of each track in the artist profile
    """

    spotify_client = SpotifyClient()

    artist_tracks = []

    artist_response = spotify_client.artist_albums(
        artist_url, album_type="album,single"
    )
    if artist_response is None:
        raise ValueError("Wrong artist id")

    albums_list = artist_response["items"]
    albums_object: Dict[str, str] = {}

    # Fetch all artist albums
    while artist_response["next"]:
        response = spotify_client.next(artist_response)
        if response is None:
            break

        artist_response = response
        albums_list.extend(artist_response["items"])

    # Remove duplicate albums
    for album in albums_list:
        # return an iterable containing the string's alphanumeric characters
        alpha_numeric_filter = filter(str.isalnum, album["name"].lower())

        # join all characters into one string
        album_name = "".join(alpha_numeric_filter)

        if albums_object.get(album_name) is None:
            albums_object[album_name] = album["uri"]

    tracks_list = []
    tracks_object: Dict[str, str] = {}

    # Fetch all tracks from all albums
    for album_uri in albums_object.values():
        response = spotify_client.album_tracks(album_uri)
        if response is None:
            continue

        album_response = response
        album_tracks = album_response["items"]

        while album_response["next"]:
            album_response = spotify_client.next(album_response)
            if album_response is None:
                break

            album_tracks.extend(album_response["items"])

        tracks_list.extend(album_tracks)

    # Filter tracks to remove duplicates and songs without our artist
    for track in tracks_list:
        # ignore None tracks or tracks without uri
        if track is not None and track.get("uri") is None:
            continue

        # return an iterable containing the string's alphanumeric characters
        alphaNumericFilter = filter(str.isalnum, track["name"].lower())

        # join all characters into one string
        trackName = "".join(alphaNumericFilter)

        if tracks_object.get(trackName) is None:
            for artist in track["artists"]:
                # get artist id from url
                # https://api.spotify.com/v1/artists/1fZAAHNWdSM5gqbi9o5iEA/albums
                #
                # split string
                #  ['https:', '', 'api.spotify.com', 'v1', 'artists',
                #  '1fZAAHNWdSM5gqbi9o5iEA', 'albums']
                #
                # get second element from the end
                # '1fZAAHNWdSM5gqbi9o5iEA'
                artistId = artist_response["href"].split("/")[-2]

                # ignore tracks that are not from our artist by checking
                # the id
                if artist["id"] == artistId:
                    tracks_object[trackName] = track["uri"]

    # Create song objects from track ids
    def get_song(track_uri):
        try:
            return from_spotify_url(
                f"https://open.spotify.com/track/{track_uri.split(':')[-1]}",
                output_format,
                use_youtube,
            )
        except (LookupError, ValueError, OSError):
            return None

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        results = executor.map(get_song, tracks_object.values())

    for result in results:
        if result is not None and result.youtube_link is not None:
            artist_tracks.append(result)

    return artist_tracks


def from_saved_tracks(
    output_format: str = None, use_youtube: bool = False, threads: int = 1
) -> List[SongObject]:
    """
    Create and return list containing SongObject for every song that user has saved

    `str` `output_format` : output format of the song

    returns a `list<songObj>` containing Url's of each track in the user's saved tracks
    """

    spotify_client = SpotifyClient()

    saved_tracks_response = spotify_client.current_user_saved_tracks()
    if saved_tracks_response is None:
        raise Exception("Couldn't get saved tracks")

    saved_tracks = saved_tracks_response["items"]
    tracks = []

    # Fetch all saved tracks
    while saved_tracks_response["next"]:
        response = spotify_client.next(saved_tracks_response)
        # response is wrong, break
        if response is None:
            break

        saved_tracks_response = response
        saved_tracks.extend(saved_tracks_response["items"])

    # Remove songs  without id
    saved_tracks = [
        track
        for track in saved_tracks
        if track is not None
        and track.get("track") is not None
        and track.get("track", {}).get("id") is not None
    ]

    def get_song(track):
        try:
            return from_spotify_url(
                "https://open.spotify.com/track/" + track["track"]["id"],
                output_format,
                use_youtube,
            )
        except (LookupError, ValueError, OSError):
            return None

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        results = executor.map(get_song, saved_tracks)

    for result in results:
        if result is not None and result.youtube_link is not None:
            tracks.append(result)

    return tracks


def from_dump(data_dump: dict) -> SongObject:
    """
    Creates song object from data dump

    `dict` `data_dump` : data dump used to create song object

    returns `SongObject`
    """

    raw_track_meta = data_dump["raw_track_meta"]
    raw_album_meta = data_dump["raw_album_meta"]
    raw_artist_meta = data_dump["raw_artist_meta"]
    youtube_link = data_dump["youtube_link"]
    lyrics = data_dump["lyrics"]

    return SongObject(
        raw_track_meta, raw_album_meta, raw_artist_meta, youtube_link, lyrics
    )



from spotdl.download import ffmpeg

valid_version = """
ffmpeg version 4.3.1-2020-11-19-essentials_build-www.gyan.dev Copyright (c) 2000-2020 the FFmpeg developers
built with gcc 10.2.0 (Rev5, Built by MSYS2 project)
configuration: --enable-gpl --enable-version3 --enable-static --disable-w32threads --disable-autodetect --enable-fontconfig --enable-iconv --enable-gnutls --enable-libxml2 --enable-gmp --enable-lzma --enable-zlib --enable-libsrt --enable-libssh --enable-libzmq --enable-avisynth --enable-sdl2 --enable-libwebp --enable-libx264 --enable-libx265 --enable-libxvid --enable-libaom --enable-libopenjpeg --enable-libvpx --enable-libass --enable-libfreetype --enable-libfribidi --enable-libvidstab --enable-libvmaf --enable-libzimg --enable-amf --enable-cuda-llvm --enable-cuvid --enable-ffnvcodec --enable-nvdec --enable-nvenc --enable-d3d11va --enable-dxva2 --enable-libmfx --enable-libgme --enable-libopenmpt --enable-libopencore-amrwb --enable-libmp3lame --enable-libtheora --enable-libvo-amrwbenc --enable-libgsm --enable-libopencore-amrnb --enable-libopus --enable-libspeex --enable-libvorbis --enable-librubberband
libavutil      56. 51.100 / 56. 51.100
libavcodec     58. 91.100 / 58. 91.100
libavformat    58. 45.100 / 58. 45.100
libavdevice    58. 10.100 / 58. 10.100
libavfilter     7. 85.100 /  7. 85.100
libswscale      5.  7.100 /  5.  7.100
libswresample   3.  7.100 /  3.  7.100
libpostproc    55.  7.100 / 55.  7.100
"""

outdated_version = """
ffmpeg version 1.0.1-2020-11-19-essentials_build-www.gyan.dev Copyright (c) 2000-2020 the FFmpeg developers
built with gcc 10.2.0 (Rev5, Built by MSYS2 project)
configuration: --enable-gpl --enable-version3 --enable-static --disable-w32threads --disable-autodetect --enable-fontconfig --enable-iconv --enable-gnutls --enable-libxml2 --enable-gmp --enable-lzma --enable-zlib --enable-libsrt --enable-libssh --enable-libzmq --enable-avisynth --enable-sdl2 --enable-libwebp --enable-libx264 --enable-libx265 --enable-libxvid --enable-libaom --enable-libopenjpeg --enable-libvpx --enable-libass --enable-libfreetype --enable-libfribidi --enable-libvidstab --enable-libvmaf --enable-libzimg --enable-amf --enable-cuda-llvm --enable-cuvid --enable-ffnvcodec --enable-nvdec --enable-nvenc --enable-d3d11va --enable-dxva2 --enable-libmfx --enable-libgme --enable-libopenmpt --enable-libopencore-amrwb --enable-libmp3lame --enable-libtheora --enable-libvo-amrwbenc --enable-libgsm --enable-libopencore-amrnb --enable-libopus --enable-libspeex --enable-libvorbis --enable-librubberband
libavutil      56. 51.100 / 56. 51.100
libavcodec     58. 91.100 / 58. 91.100
libavformat    58. 45.100 / 58. 45.100
libavdevice    58. 10.100 / 58. 10.100
libavfilter     7. 85.100 /  7. 85.100
libswscale      5.  7.100 /  5.  7.100
libswresample   3.  7.100 /  3.  7.100
libpostproc    55.  7.100 / 55.  7.100
"""

nightly_version = """
ffmpeg version N-56723-ga4e518c321-static https://johnvansickle.com/ffmpeg/  Copyright (c) 2000-2021 the FFmpeg developers
built with gcc 8 (Debian 8.3.0-6)
configuration: --enable-gpl --enable-version3 --enable-static --disable-debug --disable-ffplay --disable-indev=sndio --disable-outdev=sndio --cc=gcc --enable-fontconfig --enable-frei0r --enable-gnutls --enable-gmp --enable-libgme --enable-gray --enable-libaom --enable-libfribidi --enable-libass --enable-libvmaf --enable-libfreetype --enable-libmp3lame --enable-libopencore-amrnb --enable-libopencore-amrwb --enable-libopenjpeg --enable-librubberband --enable-libsoxr --enable-libspeex --enable-libsrt --enable-libvorbis --enable-libopus --enable-libtheora --enable-libvidstab --enable-libvo-amrwbenc --enable-libvpx --enable-libwebp --enable-libx264 --enable-libx265 --enable-libxml2 --enable-libdav1d --enable-libxvid --enable-libzvbi --enable-libzimg
libavutil      56. 72.100 / 56. 72.100
libavcodec     58.135.100 / 58.135.100
libavformat    58. 77.100 / 58. 77.100
libavdevice    58. 14.100 / 58. 14.100
libavfilter     7.111.100 /  7.111.100
libswscale      5. 10.100 /  5. 10.100
libswresample   3. 10.100 /  3. 10.100
libpostproc    55. 10.100 / 55. 10.100
"""

invalid_version = """
ffmpeg version N-56723-ga4e518c321-static https://johnvansickle.com/ffmpeg/  Copyright (c) 1998-2001 the FFmpeg developers
built with gcc 8 (Debian 8.3.0-6)
configuration: --enable-gpl --enable-version3 --enable-static --disable-debug --disable-ffplay --disable-indev=sndio --disable-outdev=sndio --cc=gcc --enable-fontconfig --enable-frei0r --enable-gnutls --enable-gmp --enable-libgme --enable-gray --enable-libaom --enable-libfribidi --enable-libass --enable-libvmaf --enable-libfreetype --enable-libmp3lame --enable-libopencore-amrnb --enable-libopencore-amrwb --enable-libopenjpeg --enable-librubberband --enable-libsoxr --enable-libspeex --enable-libsrt --enable-libvorbis --enable-libopus --enable-libtheora --enable-libvidstab --enable-libvo-amrwbenc --enable-libvpx --enable-libwebp --enable-libx264 --enable-libx265 --enable-libxml2 --enable-libdav1d --enable-libxvid --enable-libzvbi --enable-libzimg
libavutil      56. 72.100 / 56. 72.100
libavcodec     58.135.100 / 58.135.100
libavformat    58. 77.100 / 58. 77.100
libavdevice    58. 14.100 / 58. 14.100
libavfilter     7.111.100 /  7.111.100
libswscale      5. 10.100 /  5. 10.100
libswresample   3. 10.100 /  3. 10.100
libpostproc    55. 10.100 / 55. 10.100
"""


def test_valid_version(fake_process):
    fake_process.register_subprocess(["ffmpeg", "-version"], stdout=valid_version)

    assert ffmpeg.has_correct_version() == True


def test_nightly_version(fake_process):
    fake_process.register_subprocess(["ffmpeg", "-version"], stdout=nightly_version)

    assert ffmpeg.has_correct_version() == True


def test_invalid_version(fake_process, capsys):
    fake_process.register_subprocess(["ffmpeg", "-version"], stdout=invalid_version)

    assert ffmpeg.has_correct_version() == False

    output, error = capsys.readouterr()
    assert "Your FFmpeg version couldn't be detected" in error


def test_outdated_version(fake_process, capsys):
    fake_process.register_subprocess(["ffmpeg", "-version"], stdout=outdated_version)

    assert ffmpeg.has_correct_version() == False

    output, error = capsys.readouterr()
    assert "Your FFmpeg installation is too old (1.0), please update to 4.2+\n" in error



from spotdl.search.song_object import SongObject  # noqa: F401
from spotdl.search.spotify_client import SpotifyClient  # noqa: F401
from spotdl.search.song_gatherer import (
    from_spotify_url,
    from_search_term,
    from_album,
    from_playlist,
    from_artist,
    from_saved_tracks,
    from_dump,
)  # noqa: F401



import requests

from typing import List
from rapidfuzz import fuzz
from bs4 import BeautifulSoup


def _match_percentage(str1: str, str2: str, score_cutoff: float = 0) -> float:
    """
    A wrapper around `rapidfuzz.fuzz.partial_ratio` to handle UTF-8 encoded
    emojis that usually cause errors

    `str` `str1` : a random sentence
    `str` `str2` : another random sentence
    `float` `score_cutoff` : minimum score required to consider it a match
        returns 0 when similarity < score_cutoff

    RETURNS `float`
    """

    # ! this will throw an error if either string contains a UTF-8 encoded emoji
    try:
        return fuzz.partial_ratio(str1, str2, score_cutoff=score_cutoff)

    # ! we build new strings that contain only alphanumerical characters and spaces
    # ! and return the partial_ratio of that
    except:  # noqa:E722
        new_str1 = ""

        for each_letter in str1:
            if each_letter.isalnum() or each_letter.isspace():
                new_str1 += each_letter

        new_str2 = ""

        for each_letter in str2:
            if each_letter.isalnum() or each_letter.isspace():
                new_str2 += each_letter

        return fuzz.partial_ratio(new_str1, new_str2, score_cutoff=score_cutoff)


def _parse_duration(duration: str) -> float:
    """
    Convert string value of time (duration: "25:36:59") to a float value of seconds (92219.0)
    """
    try:
        # {(1, "s"), (60, "m"), (3600, "h")}
        mapped_increments = zip([1, 60, 3600], reversed(duration.split(":")))
        seconds = 0
        for multiplier, time in mapped_increments:
            seconds += multiplier * int(time)

        return float(seconds)

    # ! This usually occurs when the wrong string is mistaken for the duration
    except (ValueError, TypeError, AttributeError):
        return 0.0


def _create_song_title(song_name: str, song_artists: List[str]) -> str:
    joined_artists = ", ".join(song_artists)
    return f"{joined_artists} - {song_name}".lower()


def _get_song_lyrics(song_name: str, song_artists: List[str]) -> str:
    """
    `str` `song_name` : name of song

    `list<str>` `song_artists` : list containing name of contributing artists

    RETURNS `str`: Lyrics of the song.

    Gets the metadata of the song.
    """

    headers = {
        "Authorization": "Bearer alXXDbPZtK1m2RrZ8I4k2Hn8Ahsd0Gh_o076HYvcdlBvmc0ULL1H8Z8xRlew5qaG",
    }
    api_search_url = "https://api.genius.com/search"
    search_query = f'{song_name} {", ".join(song_artists)}'

    try:
        api_response = requests.get(
            api_search_url, params={"q": search_query}, headers=headers
        ).json()

        song_id = api_response["response"]["hits"][0]["result"]["id"]
        song_api_url = f"https://api.genius.com/songs/{song_id}"

        api_response = requests.get(song_api_url, headers=headers).json()

        song_url = api_response["response"]["song"]["url"]

        genius_page = requests.get(song_url)
        soup = BeautifulSoup(genius_page.text, "html.parser")
        lyrics_div = soup.select_one("div.lyrics")

        if lyrics_div is not None:
            return lyrics_div.get_text().strip()

        return ""
    except:  # noqa: E722
        return ""
